/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "crypt.h"
#include "server.h"

extern const char *status_item_list[NUM_STAT_ITEMS];
extern struct misc_meta_data *mmd;
extern struct gengetopt_args_info conf;
extern struct dbtool dblist[];
extern struct audio_format afl[];
extern char *user_list;

static int com_si(int, int, char **);
static int com_version(int, int, char **);
static int com_sb(int, int, char **);
static int com_sc(int, int, char **);
static int com_stat(int, int, char **);
static int com_help(int, int, char **);
static int com_hup(int, int, char **);
static int com_term(int, int, char **);
static int com_play(int, int, char **);
static int com_stop(int, int, char **);
static int com_pause(int, int, char **);
static int com_next(int, int, char **);
static int com_nomore(int, int, char **);
static int com_cdt(int, int, char **);
static int com_ff(int, int, char **);
static int com_jmp(int, int, char **);


enum {
	E_GENERAL,	E_SYNTAX,	E_SEND,		E_RECV,
	E_NODBTOOL,	E_NOAUDIOFILE,	E_FORK,		E_PIPE,
	E_AUTH,		E_CMD_INVALID,	E_PERM,		E_SPLIT,
	E_RAND,		E_USERLIST,	E_NOUSER,	E_ENCRYPT
};
static const char *internal_err_list[] = {
	"general error",
	"syntax error",
	"send error",
	"receive error",

	"no such database tool",
	"no audio file",
	"can not fork",
	"pipe error",

	"authentication error",
	"invalid command",
	"permission denied",
	"can not split command line args",

	"urandom error",
	"problems reading user list",
	"you don't exist. Go away.",
	"crypt error"
};
/* commands that are handled by the server itself */
static struct command cmd_struct[] = {
	{
		.name = "cdt",
		.handler = com_cdt,
		.perms = DB_READ | DB_WRITE,
		.description = "change database tool",
		.usage = "cdt name_of_new_dbtool",
		.help =
                "Deactivate current dbtool and activate \"name_of_new_dbtool\" \n"
	},
	{
		.name = "ff",
		.handler = com_ff,
		.perms = AFS_READ | AFS_WRITE,
		.description = "jmp amount of time forwards or backwards "
			"in current song",
		.usage = "ff n[-]",
		.help =
		"Jump n seconds forward or backward.\n"
	},
	{
		.name = "help",
		.handler = com_help,
		.perms = 0,
		.description = "print help text",
		.usage = "help [command]",
		.help =
                "Without any arguments, help prints a list of        \n"
                "availible commands. When issued with a command name \n"
                "as first argument, help prints out a description for\n"
                "that command.                                       \n"
	},
	{
		.name = "hup",
		.handler = com_hup,
		.perms = AFS_WRITE,
		.description = "force reload of config file and log file",
		.usage = "hup",
		.help =
                "After rereading the config file, a signal is sent to\n"
                "all children which forces them to close/reopen the  \n"
                "log file.                                           \n"
	},
	{
		.name = "jmp",
		.handler = com_jmp,
		.perms = AFS_READ | AFS_WRITE,
		.description = "jmp to given position in current song",
		.usage = "jmp [n]",
		.help =
		"Jump to position n, where 0 <= n <= 100\n"
	},
	{
		.name = "next",
		.handler = com_next,
		.perms = AFS_READ | AFS_WRITE,
		.description = "skip rest of current song",
		.usage = "next",
		.help =
		"If stopped or paused, start playing.\n"
	},
	{
		.name = "nomore",
		.handler = com_nomore,
		.perms = AFS_READ | AFS_WRITE,
		.description = "stop playing after current song is over",
		.usage = "nomore",
		.help =
		"Use nomore instead of stop if you don't like\n"
		"sudden endings.\n"
	},
	{
		.name ="pause",
		.handler = com_pause,
		.perms = AFS_READ | AFS_WRITE,
		.description = "pause current song",
		.usage = "pause",
		.help =
		"Use play to restart.\n"
	},
	{
		.name = "play",
		.handler = com_play,
		.perms = AFS_READ | AFS_WRITE,
		.description = "start playing or resume playing when paused",
		.usage = "play [stream]",
		.help =
		"Start/continue streaming.\n"
	},
	{
		.name = "sb",
		.handler = com_sb,
		.perms = AFS_READ,
		.description = "print status bar for current song",
		.usage = "sb [n]",
		.help =
                "Without any arguments, sb continuously prints a     \n"
                "status bar of the form                              \n"

                        "\n\t\"12:34 [56:12] (56%) filename\"\n\n"

                "indicating playing time, remaining time, percentage \n"
                "and the name of the file beeing streamed.  Use the  \n"
                "optional number n to let stat exit after having     \n"
                "displayed the status bar n times.                   \n"
	},
	{
		.name = "sc",
		.handler = com_sc,
		.perms = AFS_READ,
		.description = "print name of song whenever song changes",
		.usage = "sc [n]",
		.help =
                "sc prints exactly one line (the filename of the song  \n"
                "being played) whenever the audio file changes. Stops  \n"
                "after n iterations, or never if n is not specified.\n "
	},
	{
		.name = "si",
		.handler = com_si,
		.perms = 0,
		.description = "print server info",
		.usage = "si",
		.help =
                "Print server uptime and other information.      \n"
	},
	{
		.name = "stat",
		.handler = com_stat,
		.perms = AFS_READ,
		.description = "print status info for current song",
		.usage = "stat [n]",
		.help =
                "Without any arguments, stat continuously prints     \n"
                "status messages of the song being streamed. Use the \n"
                "optional number n to let stat exit after having     \n"
                "displayed status n times.                           \n"
	},
	{
		.name = "stop",
		.handler = com_stop,
		.perms = AFS_READ | AFS_WRITE,
		.description = "stop playing",
		.usage = "stop",
		.help =
                "When stopped, server still accepts connections,     \n"
                "so you can still connect to the server and issue    \n"
                "commands, eg. play.                                 \n"
	},
	{
		.name = "term",
		.handler = com_term,
		.perms = AFS_READ | AFS_WRITE,
		.description = "terminate server and audio file sender",
		.usage = "term",
		.help =
                "Shuts down the server. Instead of this command, you \n"
                "can also send SIGINT or SIGTERM. It should never be \n"
                "necessary to send SIGKILL.                          \n"
	},
	{
		.name = "version",
		.handler = com_version,
		.perms = 0,
		.description = "print server's version",
		.usage = "version",
		.help =
		"Show version and other info\n"
	},
	/* this indicates the end of the list. Do not touch. */
	{
		.name = NULL,
	}
};

static void dummy(int s)
{}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * compute human readable string containing
 * afs_status for given integer value
 */
static char *afs_status_tohuman(unsigned int flags)
{
	if (flags & AFS_PLAYING)
		return para_strdup("playing");
	else if (flags & AFS_NEXT)
		return para_strdup("stopped");
	else
		return para_strdup("paused");
}


/*
 * return human readable permission string. Never returns NULL.
 */
char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(7 * sizeof(char));

	msg[0] = perms & DB_READ? 'd' : '-';
	msg[1] = perms & DB_WRITE? 'D' : '-';
	msg[2] = perms & AFS_READ? 'a' : '-';
	msg[3] = perms & AFS_WRITE? 'A' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *afs_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(6 * sizeof(char));

	msg[0] = (flags & AFS_PLAYING)? 'P' : '_';
	msg[1] = (flags & AFS_NOMORE)? 'O' : '_';
	msg[2] = (flags & AFS_NEXT)? 'N' : '_';
	msg[3] = (flags & AFS_REPOS)? 'R' : '_';
	msg[4] = (flags & AFS_STREAM_WRITER)? 'W' : '_';
	msg[5] = '\0';
	return msg;
}

/*
 * compute status bar string. Never returns NULL
 */
char *get_sb_string(struct misc_meta_data *nmmd)
{
	char *base, *ret;
	long long unsigned secs = 0, rsecs = 0, percent = 0;

	base = para_basename(nmmd->filename);
	if (!base)
		return para_strdup("");
	if (!base[0])
		return base;
	if (nmmd->chunks_total) {
		secs = (long long) nmmd->seconds_total * nmmd->chunks_sent
			/ nmmd->chunks_total;
		rsecs = (long long) nmmd->seconds_total *
			(nmmd->chunks_total - nmmd->chunks_sent)
			/ nmmd->chunks_total;
		percent = 100 * ((nmmd->chunks_sent + 5) / 10)
			/ ((nmmd->chunks_total + 5) / 10);
	}
	ret = make_message("%llu:%02llu [%llu:%02llu] (%llu%%) %s",
		secs / 60, secs % 60,
		rsecs / 60, rsecs % 60,
		percent,
		base
	);
	free(base);
	return ret;
}

static char *get_status(struct misc_meta_data *nmmd)
{
	char *bar, *ret, mtime[30];
	char *status, *flags; /* afs status info */
	char *ut = uptime_str(server_uptime(GET));
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval now;
	struct tm mtime_tm;

	localtime_r(&mmd->mtime, &mtime_tm);
	strftime(mtime, 29, "%a %b %d %Y", &mtime_tm);
	gettimeofday(&now, NULL);
	/* report real status */
	status = afs_status_tohuman(nmmd->afs_status_flags);
	flags = afs_get_status_flags(nmmd->afs_status_flags);
	bar = para_basename(nmmd->filename);
	ret = make_message(
		"%s:%lu\n"	"%s:%s\n"		"%s:%i\n"	"%s:%u\n"
		"%s:%s\n"	"%s:%s\n"	"%s:%s\n"	"%s:%s\n"
		"%s:%li\n" 	"%s:%s\n"	"%s"		"%s"
		"%s:%s\n"	"%s:%lu.%lu\n"	"%s:%lu.%lu\n",
		status_item_list[SI_FILE_SIZE], nmmd->size / 1024,
		status_item_list[SI_MTIME], mtime,
		status_item_list[SI_LENGTH], nmmd->seconds_total,
		status_item_list[SI_NUM_PLAYED], nmmd->num_played,

		status_item_list[SI_STATUS_BAR], bar,
		status_item_list[SI_STATUS], status,
		status_item_list[SI_STATUS_FLAGS], flags,
		status_item_list[SI_DBTOOL], dblist[nmmd->dbt_num].name,

		status_item_list[SI_OFFSET], offset,
		status_item_list[SI_FORMAT], audio_format_name(nmmd->audio_format),
		mmd->dbinfo,
		nmmd->audio_file_info,

		status_item_list[SI_UPTIME], ut,
		status_item_list[SI_STREAM_START], nmmd->stream_start.tv_sec,
			nmmd->stream_start.tv_usec,
		status_item_list[SI_CURRENT_TIME], now.tv_sec, now.tv_usec

	);
	free(bar);
	free(flags);
	free(status);
	free(ut);
	return ret;
}

/* server info */
static int com_si(int fd, int argc, char **argv)
{
	int i, ret;
	char *ut = uptime_str(server_uptime(GET));
	char *dbtools = NULL;
	struct mallinfo mi = mallinfo();

	mmd_lock();
	for (i = 0; dblist[i].name; i++) {
		dbtools = para_strcat(dbtools, dblist[i].name);
		dbtools = para_strcat(dbtools, " ");
	}
	ret = send_va_buffer(fd, "up: %s\nplayed: %u\n"
		"connections (accepted/total): %u/%u\n"
		"active connections: %u\n"
		"current loglevel: %i\n"
		"supported database tools: %s\n"
		"server pid: %d\n"
		"mallinfo: %d\n"
		"supported audio formats: %s\n",
		ut, mmd->num_played, mmd->num_commands,
		mmd->num_connects, mmd->active_connections,
		conf.loglevel_arg,
		dbtools,
		getppid(),
		mi.arena / 1024,
		afs_supported_audio_formats()
	) < 0? -E_SEND : 1;
	mmd_unlock();
	free(ut);
	if (dbtools)
		free(dbtools);
	return ret;
}

/* version */
static int com_version(int socket_fd, int argc, char **argv)
{
	return send_buffer(socket_fd, "para_server version " VERSION  " | "
			COPYRIGHT "\n" "built: " BUILD_DATE " (" SYSTEM ", "
			CC_VERSION ")\n ") < 0? -E_SEND : 1;
}

/* sc */
static int com_sc(int socket_fd, int argc, char **argv)
{
	char *name = NULL;
	int ret, old = 0, count = -1; /* print song change forever */

	if (argc)
		count = atoi(argv[1]);
repeat:
	mmd_lock();
	if (old != mmd->num_played) {
		old = mmd->num_played;
		name = para_strdup(mmd->filename);
	}
	mmd_unlock();
	if (name) {
		ret = send_va_buffer(socket_fd, "%s\n", name);
		free(name);
		name = NULL;
		if (ret < 0)
			return -E_SEND;
		if (argc && !--count)
			return 1;
	}
	usleep(500000);
	goto repeat;
}

/* sb */
static int com_sb(int socket_fd, int argc, char **argv)
{
	char *sb;
	int ret, nr = -1;	/* status bar will be printed that many
				 * times. Negative value means: print
				 * forever
				 */
	if (argc)
		nr = atoi(argv[1]);
	while (nr) {
		mmd_lock();
		sb = get_sb_string(mmd);
		mmd_unlock();
		ret = send_va_buffer(socket_fd, "%s\n", sb);
		free(sb);
		if (ret < 0)
			return -E_SEND;
		if (nr == 1)
			return 1;
		usleep(500000);
		if (nr > 0)
			nr--;
	}
	return 1;
}

/* stat */
static int com_stat(int socket_fd, int argc, char **argv)
{
//	char *old_stat = NULL, *old_dbinfo = NULL;
	int ret, num = 0;/* status will be printed that many
			  * times. num <= 0 means: print forever
			  */
	int numplayed = 0; /* used to detect song change */
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;

	signal(SIGUSR1, dummy);

	if (argc)
		num = atoi(argv[1]);
	for (;;) {

		mmd_dup(nmmd);
		s = get_status(nmmd);
		ret = send_buffer(socket_fd, s) < 0?
			-E_SEND : 1;
		free(s);
		if (ret < 0)
			goto out;
		numplayed = nmmd->num_played;
		ret = 1;
		if (num == 1)
			goto out;
		usleep(500000 * 100);
	}
out:
	return ret;
}

static int send_description(int fd, struct command *cmd, int handler, int num)
{
	int ret, i;

	for (i = 1; cmd->name && (!num || i <= num); cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = send_va_buffer(fd, "%s\t%s\t%s\t%s\n", cmd->name,
			handler? "dbtool" : "server",
			perms,
			cmd->description) < 0? -E_SEND : 1;
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

static struct command *get_cmd_ptr(char *name, int *handler, const char ***err_list)
{
	struct command *cmd = cmd_struct;

	if (handler)
		*handler = 0; /* server commands */
	if (err_list)
		*err_list = internal_err_list;
	for (cmd = cmd_struct; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	/* not found, look for dbtool commands */
	if (handler)
		*handler = 1; /* (1st) database tool */
	mmd_lock();
	if (err_list)
		*err_list = dblist[mmd->dbt_num].err_list;
	cmd = dblist[mmd->dbt_num].cmd_list;
	mmd_unlock();
	for (; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	return NULL;
}

/* help */
static int com_help(int fd, int argc, char **argv)
{
	struct command *cmd;
	char *perms;
	int ret, handler;

	if (!argc) {
		/* no argument given, print list of commands */
		if ((ret = send_description(fd, cmd_struct, 0, 0)) < 0)
			return ret;
		mmd_lock();
		cmd = dblist[mmd->dbt_num].cmd_list;
		mmd_unlock();
		return send_description(fd, cmd, 1, 0);
	}
	/* argument given for help */
	cmd = get_cmd_ptr(argv[1], &handler, NULL);
	if (!cmd)
		return send_buffer(fd, "Unknown command\n") < 0?
			-E_SEND : 1;
	perms = cmd_perms_itohuman(cmd->perms);
	ret = send_va_buffer(fd,
		"NAME:\n~~~~~\n\t%s -- %s\n\n"
		"USAGE:\n~~~~~~\n\t%s\n\n"
		"DESCRIPTION:\n~~~~~~~~~~~~\n%s\n"
		"HANDLER:\n~~~~~~~~\n"
		"This command is handled by %s\n\n"
		"PERMS:\n~~~~~~\n"
		"Needed privileges for %s: %s\n",
		argv[1],
		cmd->description,
		cmd->usage,
		cmd->help,
		handler? "dbtool" : "server",
		argv[1],
		perms
	) < 0? -E_SEND : 0;
	free(perms);
	return ret;
}

/* hup */
static int com_hup(int socket_fd, int argc, char **argv)
{
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
static int com_term(int socket_fd, int argc, char **argv)
{
	kill(getppid(), SIGTERM);
	return 1;
}

static int com_play(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_NOMORE;
	mmd_unlock();
	return 1;

}

/* stop */
static int com_stop(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->new_afs_status_flags &= ~AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_REPOS;
	mmd->new_afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 1;
}

/* pause */
static int com_pause(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->new_afs_status_flags &= ~AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 1;
}

static int com_cdt(int fd, int argc, char **argv)
{
	int i, ret;

	if (!argc) {
		char *dbtool;
		mmd_lock();
		dbtool = para_strdup(dblist[mmd->dbt_num].name);
		mmd_unlock();
		ret = send_va_buffer(fd, "%s\n", dbtool) < 0?
			-E_SEND : 1;
		free(dbtool);
		return ret;
	}
	for (i = 0; dblist[i].name; i++) {
		if (strcmp(dblist[i].name, argv[1]))
			continue;
		mmd_lock();
		mmd->dbt_change = i;
		mmd->events++;
		mmd_unlock();
		return 1;
	}
	return -E_NODBTOOL;
}

/* next */
static int com_next(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 1;
}

/* nomore */
static int com_nomore(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_NOMORE;
	mmd_unlock();
	return 1;
}

/* ff */
static int com_ff(int socket_fd, int argc, char **argv)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (!argc)
		return -E_SYNTAX;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -E_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mmd_lock();
	ret = -E_NOAUDIOFILE;
	if (!mmd->chunks_total || !mmd->seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->seconds_total;
	else
		promille += 1000 * i / mmd->seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille >  1000) {
		mmd->new_afs_status_flags |= AFS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->chunks_total * promille) / 1000;
	mmd->new_afs_status_flags |= AFS_REPOS;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mmd_unlock();
	return ret;
}

/* jmp */
static int com_jmp(int socket_fd, int argc, char **argv)
{
	long unsigned int i;
	int ret;

	if (!argc)
		return -E_SYNTAX;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -E_SYNTAX;
	mmd_lock();
	ret = -E_NOAUDIOFILE;
	if (!mmd->chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->chunks_total * i + 50)/ 100;
	PARA_INFO_LOG("sent: %lu,  offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_afs_status_flags |= AFS_REPOS;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mmd_unlock();
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, struct command *cmd_ptr)
{
	PARA_DEBUG_LOG("%s", "checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned.
 */
static struct command *parse_cmd(char *cmdstr, const char ***err_list)
{
	char buf[255];
	struct command *cmd = NULL;
	int n = 0, handler;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
	        return NULL;
	buf[n] = '\0';
	cmd = get_cmd_ptr(buf, &handler, err_list);
	if (cmd)
		PARA_DEBUG_LOG("found command %s\n", cmd->name);
	else
		PARA_ERROR_LOG("unknown command: %s\n", buf);
	return cmd;
}
/*
 * read random input from /dev/urandom into n
 */
static int read_urandom(long unsigned *n)
{
	int fd, ret = 0;

	if (!(fd = open("/dev/urandom", O_RDONLY)))
		return 0;
	if (read(fd, n, sizeof(long unsigned)))
		ret = 1;
	close(fd);
	return ret;
}




/* Open user_list file, returns pointer to opened file on success,
 * NULL on errors
 */
static FILE *open_user_list(char *file)
{
	FILE *cf;

	PARA_DEBUG_LOG("opening user list %s\n", file);
	if (!(cf = fopen(file, "r"))) {
		PARA_INFO_LOG("failed to open user list %s (%s)\n",
			file, strerror(errno));
	        return NULL;
	}
	return cf;
}

/*
 * lookup user in user_list file. Fills in a user struct containing
 * filename of the user's public key as well as the permissions of that user.
 * Returns 1 on success, 0 if user does not exist and < 0 on errors.
 */
static int get_user(struct user *user) {
	FILE *file_ptr;
	char *char_ptr;
	char line[MAXLINE];
	/* keyword, user, key, perms */
	char w[MAXLINE], n[MAXLINE], k[MAXLINE], p[MAXLINE], tmp[4][MAXLINE];
	int num;

	file_ptr = open_user_list(user_list);
	if (!file_ptr)
		return -E_USERLIST;
	while (fgets(line, MAXLINE, file_ptr)) {
//		PARA_DEBUG_LOG("%s: Read line (%i bytes) "
//			"from config file\n", __func__, strlen(line));
		if (sscanf(line,"%200s %200s %200s %200s", w, n, k, p) < 3)
			continue;
		if (!strcmp(w, "user") && !strcmp(user->name, n)) {
			PARA_DEBUG_LOG("found entry for %s\n", n);
			strcpy(user->name, n);
			strcpy(user->pubkey_file, k);
			user->perms = 0;
			char_ptr = p;
			num = sscanf(char_ptr, "%200[A-Z_],%200[A-Z_],%200[A-Z_],%200[A-Z_]",
				tmp[0], tmp[1], tmp[2], tmp[3]);
			PARA_DEBUG_LOG("found %i perm entries\n",
				num);
			user->perms = 0;
			while (num > 0) {
				num--;
				//PARA_DEBUG_LOG("%s: tmp[%i]=%s\n", __func__,
				//	num, tmp[num]);
				if (!strcmp(tmp[num], "AFS_READ"))
					user->perms =
						user->perms | AFS_READ;
				else if (!strcmp(tmp[num], "AFS_WRITE"))
					user->perms =
						user->perms | AFS_WRITE;
				else if (!strcmp(tmp[num], "DB_READ"))
					user->perms = user->perms | DB_READ;
				else if (!strcmp(tmp[num], "DB_WRITE"))
					user->perms = user->perms | DB_WRITE;
				else /* unknown permission */
					PARA_WARNING_LOG("unknown permission:"
						"%s\n", tmp[num]);
			}
			fclose(file_ptr);
			return 1;
		}
	}
	fclose(file_ptr);
	return 0;
}

int handle_connect(int fd, struct sockaddr_in *addr)
{
	int numbytes, ret, argc;
	char buf[STRINGSIZE];
	unsigned char crypt_buf[MAXLINE];
	struct user u;
	struct command *cmd = NULL;
	long unsigned challenge_nr, chall_response;
	char **argv;
	const char **err_list = internal_err_list;
	char *command = NULL;

	signal(SIGCHLD, SIG_IGN); /* nec. otherwise we get openssl zombies */
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGUSR1, SIG_IGN);
	ret = -E_RAND;
	if (!read_urandom(&challenge_nr))
		goto err_out;
	/* send Welcome message */
	ret = -E_SEND;
	if (send_va_buffer(fd, "This is para_server, version " VERSION  ".\n" ) < 0)
		goto err_out;
	/* recv auth request line */
	ret = -E_RECV;
	if ((numbytes = recv_buffer(fd, buf, MAXLINE - 1)) <= 6)
		goto err_out;
	buf[numbytes] = '\0';
	ret = -E_AUTH;
	if (strncmp(buf, "auth ", 5))
		goto err_out;
	chop(buf);
	strcpy(u.name, buf + 5); /* ok, but ugly */
	PARA_DEBUG_LOG("received auth request for user %s\n", u.name);
	ret = -E_USERLIST;
	/* lookup user in list file */
	if ((ret = get_user(&u)) < 0)
		goto err_out;
	if (!ret) { /* user not found */
		PARA_WARNING_LOG("auth request for unknown user %s\n", u.name);
		ret = -E_NOUSER;
		goto err_out;
	}
	if ((ret = para_encrypt(u.pubkey_file, challenge_nr, crypt_buf)) <= 0) {
		ret = -E_ENCRYPT;
		goto err_out;
	}
	numbytes = ret;
	PARA_DEBUG_LOG("sending %d byte challenge\n", numbytes);
	/* We can't use send_buffer here since buf may contain null bytes */
	ret = -E_SEND;
	if (send_bin_buffer(fd,(char *) crypt_buf, numbytes) < 0)
		goto err_out;
	/* recv decrypted number */
	ret = -E_RECV;
	if ((numbytes = recv_buffer(fd, buf, sizeof(buf) - 1)) <= 0)
		goto err_out;
	ret = -E_AUTH;
	buf[numbytes] = '\0';
	if (sscanf(buf, "challenge_response:%lu", &chall_response) < 1
			|| chall_response != challenge_nr) {
		PARA_WARNING_LOG("%s", "bad auth, aborting\n");
		goto err_out;
	}
	/* auth successful. Send 'Proceed' message */
	PARA_INFO_LOG("good auth for %s (%lu)\n", u.name,
		challenge_nr);
	ret = -E_SEND;
	if (send_buffer(fd, PROCEED_MSG "\n") < 0)
		goto err_out;
	/* read command */
	while ((numbytes = recv_buffer(fd, buf, sizeof(buf) - 1)) > 0) {
		char *p;
		buf[numbytes] = '\0';
		command = para_strcat(command, buf);
		if ((p = strstr(command, EOC_MSG))) {
			*p = '\0';
			break;
		}
	}
	ret = -E_RECV;
	if (numbytes < 0)
		goto err_out;
	ret = -E_CMD_INVALID;
	/* parse command */
	if (!(cmd = parse_cmd(command, &err_list))) {
		err_list = internal_err_list;
		goto err_out;
	}
	/* valid command, check permissions */
	ret = check_perms(u.perms, cmd);
	if (ret < 0) {
		/* insufficient perms */
		err_list = internal_err_list;
		PARA_NOTICE_LOG("insufficient permissions for command %s\n",
			cmd->name);
		goto err_out;
	}
	/* valid command and sufficient perms */
	ret = -E_SPLIT;
	argc = split_args(command, &argv, '\n');
	if (!argv)
		goto err_out;
	argv[0] = cmd->name;
	mmd_lock();
	mmd->num_commands++;
	mmd_unlock();
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name, u.name,
		inet_ntoa(addr->sin_addr));
	ret = cmd->handler(fd, argc, argv);
	PARA_INFO_LOG("command handler returned %s\n",
		ret >= 0?  "success" : "failure");

	if (ret >= 0) {
		ret = EXIT_SUCCESS;
		goto out;
	}
err_out:
	if (ret != -E_SEND) {
		PARA_DEBUG_LOG("sending error response #%d (%s)\n",
			-ret, err_list[-ret]);
		send_va_buffer(fd, "%s\n", err_list[-ret]);
	}
	ret = EXIT_FAILURE;
out:
	free(command);
	mmd_lock();
	if (cmd && cmd->perms & DB_WRITE)
		mmd->events++;
	mmd->active_connections--;
	mmd_unlock();
	return ret;
}
